package sun.security.util.math.intpoly;

import sun.security.util.math.MutableIntegerModuloP;
import sun.security.util.math.IntegerModuloP;
import sun.security.util.math.ImmutableIntegerModuloP;
import java.nio.ByteOrder;
import java.util.Arrays;
import java.nio.ByteBuffer;
import sun.security.util.math.SmallValue;
import java.math.BigInteger;
import sun.security.util.math.IntegerFieldModuloP;

public abstract class IntegerPolynomial implements IntegerFieldModuloP
{
    protected static final BigInteger TWO;
    protected final int numLimbs;
    private final BigInteger modulus;
    protected final int bitsPerLimb;
    private final long[] posModLimbs;
    private final int maxAdds;
    
    protected abstract void reduce(long[] p0);
    
    protected void multByInt(long[] array, long n) {
        for (int i = 0; i < array.length; ++i) {
            int n2 = i;
            array[n2] *= n;
        }
        this.reduce(array);
    }
    
    protected abstract void mult(long[] p0, long[] p1, long[] p2);
    
    protected abstract void square(long[] p0, long[] p1);
    
    IntegerPolynomial(int bitsPerLimb, int numLimbs, int maxAdds, BigInteger modulus) {
        this.numLimbs = numLimbs;
        this.modulus = modulus;
        this.bitsPerLimb = bitsPerLimb;
        this.maxAdds = maxAdds;
        this.posModLimbs = this.setPosModLimbs();
    }
    
    private long[] setPosModLimbs() {
        long[] array = new long[this.numLimbs];
        this.setLimbsValuePositive(this.modulus, array);
        return array;
    }
    
    protected int getNumLimbs() {
        return this.numLimbs;
    }
    
    public int getMaxAdds() {
        return this.maxAdds;
    }
    
    @Override
    public BigInteger getSize() {
        return this.modulus;
    }
    
    @Override
    public IntegerPolynomial.ImmutableElement get0() {
        return new IntegerPolynomial.ImmutableElement(false);
    }
    
    @Override
    public IntegerPolynomial.ImmutableElement get1() {
        return new IntegerPolynomial.ImmutableElement(true);
    }
    
    @Override
    public IntegerPolynomial.ImmutableElement getElement(BigInteger bigInteger) {
        return new IntegerPolynomial.ImmutableElement(bigInteger);
    }
    
    @Override
    public SmallValue getSmallValue(int a) {
        int i = 1 << this.bitsPerLimb - 1;
        if (Math.abs(a) >= i) {
            throw new IllegalArgumentException("max magnitude is " + i);
        }
        return new IntegerPolynomial.Limb(a);
    }
    
    protected void encode(ByteBuffer byteBuffer, int n, byte i, long[] array) {
        int n2 = (8 * n + (32 - Integer.numberOfLeadingZeros((int)i)) + this.bitsPerLimb - 1) / this.bitsPerLimb;
        if (n2 > this.numLimbs) {
            long[] array2 = new long[n2];
            this.encodeSmall(byteBuffer, n, i, array2);
            System.arraycopy(array2, 0, array, 0, array.length);
        }
        else {
            this.encodeSmall(byteBuffer, n, i, array);
        }
    }
    
    protected void encodeSmall(ByteBuffer byteBuffer, int n, byte b, long[] a) {
        int fromIndex = 0;
        long n2 = 0L;
        int n3 = 0;
        for (int i = 0; i < n; ++i) {
            long n4 = (long)(byteBuffer.get() & 0xFF);
            if (n3 + 8 >= this.bitsPerLimb) {
                int n5 = this.bitsPerLimb - n3;
                a[fromIndex++] = n2 + ((n4 & (long)(255 >> 8 - n5)) << n3);
                n2 = n4 >> n5;
                n3 = 8 - n5;
            }
            else {
                n2 += n4 << n3;
                n3 += 8;
            }
        }
        if (b != 0) {
            long n6 = (long)(b & 0xFF);
            if (n3 + 8 >= this.bitsPerLimb) {
                int n7 = this.bitsPerLimb - n3;
                a[fromIndex++] = n2 + ((n6 & (long)(255 >> 8 - n7)) << n3);
                n2 = n6 >> n7;
            }
            else {
                n2 += n6 << n3;
            }
        }
        if (fromIndex < a.length) {
            a[fromIndex++] = n2;
        }
        Arrays.fill(a, fromIndex, a.length, 0L);
        this.postEncodeCarry(a);
    }
    
    protected void encode(byte[] array, int offset, int length, byte b, long[] array2) {
        ByteBuffer wrap = ByteBuffer.wrap(array, offset, length);
        wrap.order(ByteOrder.LITTLE_ENDIAN);
        this.encode(wrap, length, b, array2);
    }
    
    protected void postEncodeCarry(long[] array) {
        this.reduce(array);
    }
    
    @Override
    public IntegerPolynomial.ImmutableElement getElement(byte[] array, int n, int n2, byte b) {
        long[] array2 = new long[this.numLimbs];
        this.encode(array, n, n2, b, array2);
        return new IntegerPolynomial.ImmutableElement(array2, 0);
    }
    
    protected BigInteger evaluate(long[] array) {
        BigInteger bigInteger = BigInteger.ZERO;
        for (int i = array.length - 1; i >= 0; --i) {
            bigInteger = bigInteger.shiftLeft(this.bitsPerLimb).add(BigInteger.valueOf(array[i]));
        }
        return bigInteger.mod(this.modulus);
    }
    
    protected long carryValue(long n) {
        return n + (long)(1 << this.bitsPerLimb - 1) >> this.bitsPerLimb;
    }
    
    protected void carry(long[] array, int i, int n) {
        long carryOut;
        int n2;
        for (; i < n; ++i) {
            carryOut = this.carryOut(array, i);
            n2 = i + 1;
            array[n2] += carryOut;
        }
    }
    
    protected void carry(long[] array) {
        this.carry(array, 0, array.length - 1);
    }
    
    protected long carryOut(long[] array, int n) {
        long carryValue = this.carryValue(array[n]);
        array[n] -= carryValue << this.bitsPerLimb;
        return carryValue;
    }
    
    private void setLimbsValue(BigInteger bigInteger, long[] array) {
        this.setLimbsValuePositive(bigInteger, array);
        this.carry(array);
    }
    
    protected void setLimbsValuePositive(BigInteger shiftRight, long[] array) {
        BigInteger value = BigInteger.valueOf((long)(1 << this.bitsPerLimb));
        for (int i = 0; i < array.length; ++i) {
            array[i] = shiftRight.mod(value).longValue();
            shiftRight = shiftRight.shiftRight(this.bitsPerLimb);
        }
    }
    
    protected abstract void finalCarryReduceLast(long[] p0);
    
    protected void finalReduce(long[] array) {
        long n;
        int n4;
        for (int i = 0; i < 2; n4 = this.numLimbs - 1, array[n4] += n, ++i) {
            this.finalCarryReduceLast(array);
            n = 0L;
            for (int j = 0; j < this.numLimbs - 1; ++j) {
                int n2 = j;
                array[n2] += n;
                n = array[j] >> this.bitsPerLimb;
                int n3 = j;
                array[n3] -= n << this.bitsPerLimb;
            }
        }
        int n5 = 1;
        long[] array2 = new long[this.numLimbs];
        for (int k = this.numLimbs - 1; k >= 0; --k) {
            array2[k] = array[k] - this.posModLimbs[k];
            n5 *= (int)(array2[k] >> 63) + 1;
        }
        conditionalSwap(n5, array, array2);
    }
    
    protected void decode(long[] array, byte[] array2, int n, int n2) {
        int n3 = 0;
        long n4 = array[n3++];
        int n5 = 0;
        for (int i = 0; i < n2; ++i) {
            int n6 = i + n;
            if (n5 + 8 >= this.bitsPerLimb) {
                array2[n6] = (byte)(int)n4;
                long n7 = 0L;
                if (n3 < array.length) {
                    n7 = array[n3++];
                }
                int n8 = this.bitsPerLimb - n5;
                int n9 = 8 - n8;
                int n10 = n6;
                array2[n10] = (byte)(int)((long)array2[n10] + ((n7 & (long)(255 >> n8)) << n8));
                n4 = n7 >> n9;
                n5 = n9;
            }
            else {
                array2[n6] = (byte)(int)n4;
                n4 >>= 8;
                n5 += 8;
            }
        }
    }
    
    protected void addLimbs(long[] array, long[] array2, long[] array3) {
        for (int i = 0; i < array3.length; ++i) {
            array3[i] = array[i] + array2[i];
        }
    }
    
    protected static void conditionalAssign(int n, long[] array, long[] array2) {
        int n2 = 0 - n;
        for (int i = 0; i < array.length; ++i) {
            array[i] ^= ((long)n2 & (array[i] ^ array2[i]));
        }
    }
    
    protected static void conditionalSwap(int n, long[] array, long[] array2) {
        int n2 = 0 - n;
        for (int i = 0; i < array.length; ++i) {
            long n3 = (long)n2 & (array[i] ^ array2[i]);
            array[i] ^= n3;
            array2[i] ^= n3;
        }
    }
    
    protected void limbsToByteArray(long[] array, byte[] array2) {
        long[] array3 = (long[])array.clone();
        this.finalReduce(array3);
        this.decode(array3, array2, 0, array2.length);
    }
    
    protected void addLimbsModPowerTwo(long[] array, long[] array2, byte[] array3) {
        long[] array4 = (long[])array2.clone();
        long[] array5 = (long[])array.clone();
        this.finalReduce(array4);
        this.finalReduce(array5);
        this.addLimbs(array5, array4, array5);
        long n = 0L;
        for (int i = 0; i < this.numLimbs; ++i) {
            int n2 = i;
            array5[n2] += n;
            n = array5[i] >> this.bitsPerLimb;
            int n3 = i;
            array5[n3] -= n << this.bitsPerLimb;
        }
        this.decode(array5, array3, 0, array3.length);
    }
    
    static void access$000(IntegerPolynomial integerPolynomial, BigInteger bigInteger, long[] array) {
        integerPolynomial.setLimbsValue(bigInteger, array);
    }
    
    static {
        TWO = BigInteger.valueOf(2L);
    }
    
    private abstract class Element implements IntegerModuloP
    {
        protected long[] limbs;
        protected int numAdds;
        
        public Element(BigInteger value) {
            this.limbs = new long[IntegerPolynomial.this.numLimbs];
            this.setValue(value);
        }
        
        public Element(boolean b) {
            (this.limbs = new long[IntegerPolynomial.this.numLimbs])[0] = b;
            this.numAdds = 0;
        }
        
        private Element(long[] limbs, int numAdds) {
            this.limbs = limbs;
            this.numAdds = numAdds;
        }
        
        private void setValue(BigInteger bigInteger) {
            IntegerPolynomial.access$000(IntegerPolynomial.this, bigInteger, this.limbs);
            this.numAdds = 0;
        }
        
        @Override
        public IntegerFieldModuloP getField() {
            return IntegerPolynomial.this;
        }
        
        @Override
        public BigInteger asBigInteger() {
            return IntegerPolynomial.this.evaluate(this.limbs);
        }
        
        @Override
        public IntegerPolynomial.MutableElement mutable() {
            return new IntegerPolynomial.MutableElement((long[])this.limbs.clone(), this.numAdds);
        }
        
        protected boolean isSummand() {
            return this.numAdds < IntegerPolynomial.this.maxAdds;
        }
        
        @Override
        public IntegerPolynomial.ImmutableElement add(IntegerModuloP integerModuloP) {
            IntegerPolynomial.Element element = (IntegerPolynomial.Element)integerModuloP;
            if (!this.isSummand() || !element.isSummand()) {
                throw new ArithmeticException("Not a valid summand");
            }
            long[] array = new long[this.limbs.length];
            for (int i = 0; i < this.limbs.length; ++i) {
                array[i] = this.limbs[i] + element.limbs[i];
            }
            return new IntegerPolynomial.ImmutableElement(array, Math.max(this.numAdds, element.numAdds) + 1);
        }
        
        @Override
        public IntegerPolynomial.ImmutableElement additiveInverse() {
            long[] array = new long[this.limbs.length];
            for (int i = 0; i < this.limbs.length; ++i) {
                array[i] = -this.limbs[i];
            }
            return new IntegerPolynomial.ImmutableElement(array, this.numAdds);
        }
        
        protected long[] cloneLow(long[] array) {
            long[] array2 = new long[IntegerPolynomial.this.numLimbs];
            this.copyLow(array, array2);
            return array2;
        }
        
        protected void copyLow(long[] array, long[] array2) {
            System.arraycopy(array, 0, array2, 0, array2.length);
        }
        
        @Override
        public IntegerPolynomial.ImmutableElement multiply(IntegerModuloP integerModuloP) {
            IntegerPolynomial.Element element = (IntegerPolynomial.Element)integerModuloP;
            long[] array = new long[this.limbs.length];
            IntegerPolynomial.this.mult(this.limbs, element.limbs, array);
            return new IntegerPolynomial.ImmutableElement(array, 0);
        }
        
        @Override
        public IntegerPolynomial.ImmutableElement square() {
            long[] array = new long[this.limbs.length];
            IntegerPolynomial.this.square(this.limbs, array);
            return new IntegerPolynomial.ImmutableElement(array, 0);
        }
        
        @Override
        public void addModPowerTwo(IntegerModuloP integerModuloP, byte[] array) {
            IntegerPolynomial.Element element = (IntegerPolynomial.Element)integerModuloP;
            if (!this.isSummand() || !element.isSummand()) {
                throw new ArithmeticException("Not a valid summand");
            }
            IntegerPolynomial.this.addLimbsModPowerTwo(this.limbs, element.limbs, array);
            return;
        }
        
        @Override
        public void asByteArray(byte[] array) {
            if (!this.isSummand()) {
                throw new ArithmeticException("Not a valid summand");
            }
            IntegerPolynomial.this.limbsToByteArray(this.limbs, array);
            return;
        }
    }
    
    protected class MutableElement extends IntegerPolynomial.Element implements MutableIntegerModuloP
    {
        protected MutableElement(long[] array, int n) {
            super(array, n);
        }
        
        @Override
        public IntegerPolynomial.ImmutableElement fixed() {
            return new IntegerPolynomial.ImmutableElement((long[])this.limbs.clone(), this.numAdds);
        }
        
        @Override
        public void conditionalSet(IntegerModuloP integerModuloP, int n) {
            IntegerPolynomial.Element element = (IntegerPolynomial.Element)integerModuloP;
            IntegerPolynomial.conditionalAssign(n, this.limbs, element.limbs);
            this.numAdds = element.numAdds;
        }
        
        @Override
        public void conditionalSwapWith(MutableIntegerModuloP mutableIntegerModuloP, int n) {
            IntegerPolynomial.MutableElement mutableElement = (IntegerPolynomial.MutableElement)mutableIntegerModuloP;
            IntegerPolynomial.conditionalSwap(n, this.limbs, mutableElement.limbs);
            int numAdds = this.numAdds;
            this.numAdds = mutableElement.numAdds;
            mutableElement.numAdds = numAdds;
        }
        
        @Override
        public IntegerPolynomial.MutableElement setValue(IntegerModuloP integerModuloP) {
            IntegerPolynomial.Element element = (IntegerPolynomial.Element)integerModuloP;
            System.arraycopy(element.limbs, 0, this.limbs, 0, element.limbs.length);
            this.numAdds = element.numAdds;
            return this;
        }
        
        @Override
        public IntegerPolynomial.MutableElement setValue(byte[] array, int n, int n2, byte b) {
            IntegerPolynomial.this.encode(array, n, n2, b, this.limbs);
            this.numAdds = 0;
            return this;
        }
        
        @Override
        public IntegerPolynomial.MutableElement setValue(ByteBuffer byteBuffer, int n, byte b) {
            IntegerPolynomial.this.encode(byteBuffer, n, b, this.limbs);
            this.numAdds = 0;
            return this;
        }
        
        @Override
        public IntegerPolynomial.MutableElement setProduct(IntegerModuloP integerModuloP) {
            IntegerPolynomial.this.mult(this.limbs, ((IntegerPolynomial.Element)integerModuloP).limbs, this.limbs);
            this.numAdds = 0;
            return this;
        }
        
        @Override
        public IntegerPolynomial.MutableElement setProduct(SmallValue smallValue) {
            IntegerPolynomial.this.multByInt(this.limbs, (long)((IntegerPolynomial.Limb)smallValue).value);
            this.numAdds = 0;
            return this;
        }
        
        @Override
        public IntegerPolynomial.MutableElement setSum(IntegerModuloP integerModuloP) {
            IntegerPolynomial.Element element = (IntegerPolynomial.Element)integerModuloP;
            if (!this.isSummand() || !element.isSummand()) {
                throw new ArithmeticException("Not a valid summand");
            }
            for (int i = 0; i < this.limbs.length; ++i) {
                this.limbs[i] += element.limbs[i];
            }
            this.numAdds = Math.max(this.numAdds, element.numAdds) + 1;
            return this;
        }
        
        @Override
        public IntegerPolynomial.MutableElement setDifference(IntegerModuloP integerModuloP) {
            IntegerPolynomial.Element element = (IntegerPolynomial.Element)integerModuloP;
            if (!this.isSummand() || !element.isSummand()) {
                throw new ArithmeticException("Not a valid summand");
            }
            for (int i = 0; i < this.limbs.length; ++i) {
                this.limbs[i] -= element.limbs[i];
            }
            this.numAdds = Math.max(this.numAdds, element.numAdds) + 1;
            return this;
        }
        
        @Override
        public IntegerPolynomial.MutableElement setSquare() {
            IntegerPolynomial.this.square(this.limbs, this.limbs);
            this.numAdds = 0;
            return this;
        }
        
        @Override
        public IntegerPolynomial.MutableElement setAdditiveInverse() {
            for (int i = 0; i < this.limbs.length; ++i) {
                this.limbs[i] = -this.limbs[i];
            }
            return this;
        }
        
        public IntegerPolynomial.MutableElement setReduced() {
            IntegerPolynomial.this.reduce(this.limbs);
            this.numAdds = 0;
            return this;
        }
    }
    
    class ImmutableElement extends IntegerPolynomial.Element implements ImmutableIntegerModuloP
    {
        protected ImmutableElement(BigInteger bigInteger) {
            super(bigInteger);
        }
        
        protected ImmutableElement(boolean b) {
            super(b);
        }
        
        protected ImmutableElement(long[] array, int n) {
            super(array, n);
        }
        
        @Override
        public IntegerPolynomial.ImmutableElement fixed() {
            return this;
        }
    }
    
    class Limb implements SmallValue
    {
        int value;
        
        Limb(int value) {
            this.value = value;
        }
    }
}
